#include <asm/system.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/delay.h>
#include "../block/blk.h"
#include "scsi.h"
#include "hosts.h"
#include "wd33c93.h"

#undef NDEBUG
#define NDEBUG 0

inline unsigned char read_WD33C93(wd33c93_regs *regp,
				       unsigned char reg_num)
{
    regp->SASR = reg_num;
    return(regp->SCMD);
}
inline void write_WD33C93(wd33c93_regs *regp,
			       unsigned char reg_num, unsigned char value)
{
    regp->SASR = reg_num;
    regp->SCMD = value;
}
inline void write_WD33C93_count(wd33c93_regs *regp,
				 unsigned long value)
{
    regp->SASR = WD_TRANSFER_COUNT_MSB;
    regp->SCMD = value	>> 16;
    regp->SCMD = value >> 8;
    regp->SCMD = value;
}
inline unsigned long read_WD33C93_count(wd33c93_regs *regp)
{
    unsigned long value;

    regp->SASR = WD_TRANSFER_COUNT_MSB;
    value = regp->SCMD << 16;
    value |= regp->SCMD << 8;
    value |= regp->SCMD;

    return value;
}
inline void write_WD33C93_cmd (wd33c93_regs *regp, unsigned char *cdb)
{
    int len = COMMAND_SIZE(cdb[0]);

    regp->SASR = WD_CDB_1;
    while (len--)
	regp->SCMD = *cdb++;
}

void reset_WD33C93(struct Scsi_Host *instance)
{
    struct WD33C93_hostdata *hostdata = (struct WD33C93_hostdata *)
	    instance->hostdata;

    write_WD33C93(hostdata->regp, WD_OWN_ID,  hostdata->clock_freq | 0x0F);
    write_WD33C93(hostdata->regp, WD_CONTROL, 0x08);
    write_WD33C93(hostdata->regp, WD_COMMAND, WD_CMD_RESET);

    while (!(read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS) & ASR_INT))
	;
    read_WD33C93(hostdata->regp,  WD_SCSI_STATUS);
    write_WD33C93(hostdata->regp, WD_TIMEOUT_PERIOD, 0x16);
    write_WD33C93(hostdata->regp, WD_CONTROL,        0x08);
}

const char *wd33c93_info (void)
{
    static const char string[]="";
    return string;
}

/*
 * Function : void initialize_SCp(Scsi_Cmnd *cmd)
 *
 * Purpose : initialize the saved data pointers for cmd to point to the
 *	start of the buffer.
 *
 * Inputs : cmd - Scsi_Cmnd structure to have pointers reset.
 */

static __inline__ void initialize_SCp(Scsi_Cmnd *cmd) {
    /*
     * Initialize the Scsi Pointer field so that all of the commands in the
     * various queues are valid.
     */

    if (cmd->use_sg) {
	cmd->SCp.buffer = (struct scatterlist *) cmd->buffer;
	cmd->SCp.buffers_residual = cmd->use_sg - 1;
	cmd->SCp.ptr = (char *) cmd->SCp.buffer->address;
	cmd->SCp.this_residual = cmd->SCp.buffer->length;
    } else {
	cmd->SCp.buffer = NULL;
	cmd->SCp.buffers_residual = 0;
	cmd->SCp.ptr = (char *) cmd->request_buffer;
	cmd->SCp.this_residual = cmd->request_bufflen;
    }
}

/*
 * This routine must be called with interrupts disabled
 */
static void WD33C93_issue (struct Scsi_Host *instance)
{
    struct WD33C93_hostdata *hostdata = (struct WD33C93_hostdata *)
	    instance->hostdata;
    Scsi_Cmnd *cmd, *prev;
    int dir_in;

    /*
     * Search through the issue_queue for a command destined
     * for a target that's not busy.
     */
    for (cmd = (Scsi_Cmnd *) hostdata->issue_queue,
	 prev = NULL; cmd; prev = cmd, cmd = (Scsi_Cmnd *)
	 cmd->host_scribble)
	/*  When we find one, remove it from the issue queue. */
	if (!(hostdata->busy[cmd->target] & (1 << cmd->lun))) {
	    if (prev)
		prev->host_scribble = cmd->host_scribble;
	    else
		hostdata->issue_queue = (Scsi_Cmnd *) cmd->host_scribble;
	    cmd->host_scribble = NULL;

	    /*
	     * Start a command
	     */
#if (NDEBUG & (NDEBUG_MAIN | NDEBUG_QUEUES))
	    printk("scsi%d : WD33C93_issue() : command for target %d lun %d removed from issue_queue\n",
		   instance->host_no, cmd->target, cmd->lun);
#endif
#ifdef SCSI2
	    if (!scsi_devices[cmd->index].tagged_queue)
#endif
		hostdata->busy[cmd->target] |= (1 << cmd->lun);
	    hostdata->connected = cmd;
	    initialize_SCp(cmd);

	    dir_in = !(cmd->cmnd[0] == WRITE_6 || cmd->cmnd[0] == WRITE_10);
	    if (!dir_in)
		write_WD33C93(hostdata->regp, WD_DESTINATION_ID,
			      cmd->target);
	    else
		write_WD33C93(hostdata->regp, WD_DESTINATION_ID,
			      cmd->target | 0x40);

	    write_WD33C93(hostdata->regp, WD_TARGET_LUN, cmd->lun);
	    write_WD33C93_cmd(hostdata->regp, cmd->cmnd);

	    /* if wrong alignment/address for DMA, do PIO */
	    if ((unsigned long)cmd->SCp.ptr & hostdata->xfermask)
		hostdata->pio = 1;
	    else
		hostdata->pio = 0;

#if 1
	    /* temporarily FORCE programmed I/O (no DMA) for writes */
	    if (!dir_in)
		    hostdata->pio = 1;
#endif

	    if (hostdata->pio)
	    {
		write_WD33C93(hostdata->regp, WD_CONTROL,        0x08);
		write_WD33C93_count(hostdata->regp, 0);
	    }
	    else
	    {
		/* start dma */
		write_WD33C93(hostdata->regp, WD_CONTROL,        0x88);
		hostdata->dma_setup (cmd, dir_in);
		write_WD33C93_count(hostdata->regp, cmd->request_bufflen);
	    }
#if (NDEBUG & NDEBUG_QUEUES)
	    printk("scsi%d : set count to %d\n", instance->host_no,
		   cmd->request_bufflen);
#endif
	    write_WD33C93(hostdata->regp, WD_COMMAND, WD_CMD_SEL_XFER);
	}
}

int wd33c93_queuecommand (Scsi_Cmnd *cmd, void (*done)(Scsi_Cmnd *))
{
    struct WD33C93_hostdata *hostdata = (struct WD33C93_hostdata *)
	cmd->host->hostdata;
    struct Scsi_Host *instance = cmd->host;
    Scsi_Cmnd *tmp;
    unsigned long flags;

    /*
     * We use the host_scribble field as a pointer to the next command
     * in a queue
     */

    cmd->host_scribble = NULL;
    cmd->scsi_done = done;
    cmd->result = 0;


    /*
     * Insert the cmd into the issue queue. Note that REQUEST SENSE
     * commands are added to the head of the queue since any command will
     * clear the contingent allegience condition that exists and the
     * sense data is only guaranteed to be valid while the condition exists.
     */

    save_flags (flags);
    cli();
    if (!(hostdata->issue_queue) || (cmd->cmnd[0] == REQUEST_SENSE)) {
	cmd->host_scribble = (unsigned char *) hostdata->issue_queue;
	hostdata->issue_queue = cmd;
    } else {
	for (tmp = (Scsi_Cmnd *) hostdata->issue_queue; tmp->host_scribble;
		tmp = (Scsi_Cmnd *) tmp->host_scribble);
	tmp->host_scribble = (unsigned char *) cmd;
    }
#if (NDEBUG & NDEBUG_QUEUES)
    printk("scsi%d : command added to %s of queue\n", instance->host_no,
	(cmd->cmnd[0] == REQUEST_SENSE) ? "head" : "tail");
#endif

/* issue a command if one isn't already running. */
    if (!hostdata->running && !hostdata->connected) {
	 hostdata->running = 1;
	 WD33C93_issue(instance);
	 hostdata->running = 0;
    }
    restore_flags(flags);
    return 0;
}

void WD33C93_intr (struct Scsi_Host *instance)
{
    struct WD33C93_hostdata *hostdata = (struct WD33C93_hostdata *)
	    instance->hostdata;
    Scsi_Cmnd *cmd = (Scsi_Cmnd *)hostdata->connected;
    wd33c93_regs *regp = hostdata->regp;
    unsigned char asr, sr, phs, id;
    unsigned long flags;

    asr = read_WD33C93(regp, WD_AUXILIARY_STATUS);
    if (!(asr & ASR_INT))
	return;

    save_flags (flags);
    cli();

    hostdata->running = 1;

    sr = read_WD33C93(regp, WD_SCSI_STATUS);
    phs = read_WD33C93(regp, WD_COMMAND_PHASE);
#if (NDEBUG & NDEBUG_INTR)
    printk("scsi%d : INTERRUPT A:0x%x, P:0x%x, S:0x%x\n", instance->host_no,
	   asr,phs,sr);
#endif

redo:
    if (sr == CSR_DISC)
    {
	/*
	 * target disconnected; with queued commands, this
	 * is where we'd issue another command.
	 */
	printk ("scsi%d: target disconnected\n", instance->host_no);
	cmd->result = DID_ERROR;
	cmd->scsi_done(cmd);

	hostdata->connected = NULL;
	hostdata->busy[cmd->target] &= ~(1 << cmd->lun);
    }
    else if ((sr == (CSR_UNEXP|PHS_DATA_OUT) ||
	      sr == (CSR_UNEXP|PHS_DATA_IN))
	     && (phs & 0xf0) == 0x30 && hostdata->pio)
    {
	/*
	 * We are using PIO, so we set the count to 0 so that
	 * an interrupt would be generated when data was ready
	 * to be transferred.
	 * Perform an XFER_INFO command.
	 */
#if (NDEBUG & NDEBUG_INTR)
	printk ("scsi%d: target wants data transfer\n", instance->host_no);
#endif
	/* setup count and appropriate direction */
	write_WD33C93_count(hostdata->regp, cmd->request_bufflen);
	id = read_WD33C93(regp, WD_DESTINATION_ID);
	if (sr == 0x48)
	    id &= 0x40;
	else
	    id |= 0x40;
	write_WD33C93(regp, WD_DESTINATION_ID, id);

	/* issue the command */
	write_WD33C93(regp, WD_COMMAND,     WD_CMD_TRANS_INFO);

	restore_flags(flags);
	/* read/write data from scsi chip */
	do {
	    cli();
	    asr = read_WD33C93(regp, WD_AUXILIARY_STATUS);
	    if (asr & ASR_DBR)
	    {
		if (sr == 0x49)
		    *(cmd->SCp.ptr)++ = read_WD33C93(regp, WD_DATA);
		else
		    write_WD33C93(regp, WD_DATA, *(cmd->SCp.ptr)++);
		cmd->SCp.this_residual--;
	    }
	    restore_flags(flags);
	    /* if this command is aborted, hostdata->running will be 0 */
	} while (hostdata->running && !(asr & ASR_INT));
	cli();

	/* if command wasn't aborted! */
	if (hostdata->running)
	{
	    sr = read_WD33C93(regp, WD_SCSI_STATUS);
	    phs = read_WD33C93(regp, WD_COMMAND_PHASE);
#if (NDEBUG & NDEBUG_INTR)
	    printk ("scsi%d: asr=%x sr=%x phs=%x after data transfer\n",
		    instance->host_no, asr, sr, phs);
#endif
	    goto redo;
	}
	else
	    printk ("scsi%d: command aborted during data transfer\n",
		    instance->host_no);
    }
    else if (sr == (CSR_UNEXP|PHS_DATA_OUT) && (phs & 0xf0) == 0x30 &&
	     !(read_WD33C93(regp, WD_DESTINATION_ID) & 0x40))
    {
#if (NDEBUG & NDEBUG_INTR)
	printk("scsi%d: changing data dir..\n", instance->host_no);
#endif
	id = read_WD33C93(regp, WD_DESTINATION_ID);
	id &= ~0x40;
	write_WD33C93(regp, WD_DESTINATION_ID, id);
	if (!hostdata->pio)
	{
	    hostdata->dma_stop ((Scsi_Cmnd *)hostdata->connected);
	    hostdata->dma_setup ((Scsi_Cmnd *)hostdata->connected, 0);
	}
	write_WD33C93(regp, WD_COMMAND_PHASE, 0x41);
	write_WD33C93(regp, WD_COMMAND,       WD_CMD_SEL_XFER);
    }
    else if (sr == (CSR_UNEXP|PHS_STATUS) ||
	     sr == (CSR_XFER_DONE|PHS_STATUS))
    {
#if (NDEBUG & NDEBUG_INTR)
	printk("scsi%d: Resuming...\n", instance->host_no);
#endif
	write_WD33C93_count (regp, 0);
	if (!hostdata->pio)
	    hostdata->dma_stop (cmd);
	write_WD33C93(regp, WD_COMMAND_PHASE, 0x41);
	write_WD33C93(regp, WD_COMMAND,       WD_CMD_SEL_XFER);
    }
    else if (sr == CSR_TIMEOUT)
    {
	if (!hostdata->pio)
	    hostdata->dma_stop (cmd);
	cmd->result = DID_NO_CONNECT;
	cmd->scsi_done(cmd);

	hostdata->connected = NULL;
	hostdata->busy[cmd->target] &= ~(1 << cmd->lun);
    }
    else if (sr == CSR_SEL_XFER_DONE)
    {
	cmd->result = read_WD33C93(regp, WD_TARGET_LUN);
#if (NDEBUG & NDEBUG_INTR)
	printk("scsi%d : command for target %d, lun %d completed\n",
	       instance->host_no, cmd->target, cmd->lun);
#endif
	if (!hostdata->pio)
	    hostdata->dma_stop (cmd);
	cmd->scsi_done(cmd);

	hostdata->busy[cmd->target] &= ~(1 << cmd->lun);
	hostdata->connected = NULL;
    }
    else
    {
#if (NDEBUG & NDEBUG_INTR)
	printk("scsi%d : command for target %d, lun %d completed badly\n",
	       instance->host_no, cmd->target, cmd->lun);
#endif
	if (!hostdata->pio)
	    hostdata->dma_stop (cmd);
	cmd->result = (cmd->result & 0x00ffff) | (DID_ERROR << 16);
	cmd->scsi_done(cmd);

	hostdata->busy[cmd->target] &= ~(1 << cmd->lun);
	hostdata->connected = NULL;
    }

    if (!hostdata->connected && hostdata->issue_queue) {
#if (NDEBUG & NDEBUG_INTR)
	printk ("scsi%d: issuing command from interrupt\n",
		instance->host_no);
#endif
	WD33C93_issue (instance);
    }

    hostdata->running = 0;
    restore_flags(flags);
}

int wd33c93_abort (Scsi_Cmnd *cmd, int code)
{
    struct Scsi_Host *instance = cmd->host;
    struct WD33C93_hostdata *hostdata = (struct WD33C93_hostdata *)
	    instance->hostdata;
    Scsi_Cmnd *tmp, **prev;
    unsigned long flags;

    printk ("scsi%d: abort (%d)\n", instance->host_no, code);

    save_flags (flags);
    cli();

/*
 * Case 1 : If the command hasn't been issued yet, we simply remove it
 *	    from the issue queue.
 */
    for (prev = (Scsi_Cmnd **) &(hostdata->issue_queue),
	tmp = (Scsi_Cmnd *) hostdata->issue_queue;
	tmp; prev = (Scsi_Cmnd **) &(tmp->host_scribble), tmp =
	(Scsi_Cmnd *) tmp->host_scribble)
	if (cmd == tmp) {
	    (*prev) = (Scsi_Cmnd *) tmp->host_scribble;
	    tmp->host_scribble = NULL;
	    tmp->result = (code ? code : DID_ABORT) << 16;
	    restore_flags(flags);
	    printk("scsi%d : abort removed command from issue queue.\n",
		   instance->host_no);
	    tmp->scsi_done(tmp);
	    return 0;
	}

/*
 * Case 2 : If any commands are connected, we're going to fail the abort
 *	    and let the high level SCSI driver retry at a later time or
 *	    issue a reset.
 *
 *	    Timeouts, and therefore aborted commands, will be highly unlikely
 *	    and handling them cleanly in this situation would make the common
 *	    case of noresets less efficient, and would pollute our code.  So,
 *	    we fail.
 */

    if (hostdata->connected == cmd) {
	unsigned char sr, asr;
	unsigned long timeout;

	printk("scsi%d : aborting connected command.\n", instance->host_no);

	if (!hostdata->pio)
	    hostdata->dma_stop(cmd);

	printk("scsi%d : WD33C93 asr is %x\n", instance->host_no,
	       read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS));

	write_WD33C93(hostdata->regp, WD_COMMAND, WD_CMD_ABORT);
	timeout = 1000000;
	do {
	    asr = read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS);
	    if (asr & ASR_DBR)
		read_WD33C93(hostdata->regp, WD_DATA);
	} while (!(asr & ASR_INT) && timeout-- > 0);
	sr = read_WD33C93(hostdata->regp, WD_SCSI_STATUS);
	printk("scsi%d : WD33C93 sr is %x\n", instance->host_no,
	       read_WD33C93(hostdata->regp, WD_SCSI_STATUS));

	if (sr >= (CSR_ABORT|PHS_DATA_OUT) &&
	    sr <= (CSR_ABORT|PHS_MESS_IN))
	{
	    /*
	     * Abort command processed.
	     * Still connected.
	     * We must disconnect.
	     */
	    printk("scsi%d : count was %ld\n", instance->host_no,
		   read_WD33C93_count(hostdata->regp));

	    timeout = 1000000;
	    while ((asr & ASR_CIP) && timeout-- > 0)
		asr = read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS);
	    write_WD33C93(hostdata->regp, WD_COMMAND, WD_CMD_DISCONNECT);
	    asr = read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS);
	    if (asr & ASR_LCI)
		printk ("scsi%d: disconnect command ignored\n",
			instance->host_no);
	    timeout = 1000000;
	    while (asr & ASR_CIP && timeout-- > 0)
		asr = read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS);
	}
	asr = read_WD33C93(hostdata->regp, WD_AUXILIARY_STATUS);
	sr = read_WD33C93(hostdata->regp, WD_SCSI_STATUS);
	printk("scsi%d : asr is %x, sr is %x\n", instance->host_no, asr,
	       read_WD33C93(hostdata->regp, WD_SCSI_STATUS));

	cmd->result = (code ? code : DID_ABORT) << 16;
	cmd->scsi_done(cmd);
	hostdata->busy[cmd->target] &= ~(1 << cmd->lun);
	hostdata->connected = NULL;
	hostdata->running = 0;
#if (NDEBUG & NDEBUG_ABORT)
	printk("scsi%d : aborted connected command.\n", instance->host_no);
#endif
	WD33C93_issue (instance);
	restore_flags(flags);
	return 0;
    }

/*
 * Case 3: If the command is currently disconnected from the bus, and
 *	there are no connected commands, we reconnect the I_T_L or
 *	I_T_L_Q nexus associated with it, go into message out, and send
 *	an abort message.
 *
 * This case is especially ugly. In order to resetablish the nexus, we
 * need to call WD33C93_select().  The easiest way to implement this
 * function was to abort if the bus was busy, and let the interrupt
 * handler triggered on the SEL for reselect take care of lost arbitrations
 * where necessary, meaning interrupts need to be enabled.
 *
 * When interrupts are enabled, the queues may change - so we
 * can't remove it from the disconnected queue before selecting it
 * because that could cause a failure in hashing the nexus if that
 * device reselected.
 *
 * Since the queues may change, we can't use the pointers from when we
 * first locate it.
 *
 * So, we must first locate the command, and if WD33C93_select()
 * succeeds, then issue the abort, relocate the command and remove
 * it from the disconnected queue.
 */

    for (tmp = (Scsi_Cmnd *) hostdata->disconnected_queue; tmp;
	tmp = (Scsi_Cmnd *) tmp->host_scribble)
	if (cmd == tmp) {
	    restore_flags(flags);
#if (NDEBUG & NDEBUG_ABORT)
    printk("scsi%d : aborting disconnected command.\n", instance->host_no);
#endif

#if 0
	    if (WD33C93_select (instance, cmd, (int) cmd->tag))
		return 1;

#if (NDEBUG & NDEBUG_ABORT)
    printk("scsi%d : nexus restablished.\n", instance->host_no);
#endif

	    msg = ABORT;
	    msgptr = &msg;
	    len = 1;
	    phase = PHASE_MSGOUT;
	    WD33C93_write(INITIATOR_COMMAND_REG, ICR_BASE | ICR_ASSERT_ATN);
	    WD33C93_transfer_pio (instance, &phase, &len, &msgptr);
#endif
	    cli();
	    for (prev = (Scsi_Cmnd **) &(hostdata->disconnected_queue),
		tmp = (Scsi_Cmnd *) hostdata->disconnected_queue;
		tmp; prev = (Scsi_Cmnd **) &(tmp->host_scribble), tmp =
		(Scsi_Cmnd *) tmp->host_scribble)
		    if (cmd == tmp) {
		    *prev = (Scsi_Cmnd *) tmp->host_scribble;
		    tmp->host_scribble = NULL;
		    tmp->result = (code ? code : DID_ABORT) << 16;
		    restore_flags(flags);
		    tmp->scsi_done(tmp);
		    return 0;
		}
	}

/*
 * Case 4 : If we reached this point, the command was not found in any of
 *	    the queues.
 *
 * We probably reached this point because of an unlikely race condition
 * between the command completing successfully and the abortion code,
 * so we won't panic, but we will notify the user in case somethign really
 * broke.
 */

    restore_flags(flags);
    printk("scsi%d : warning : SCSI command probably completed successfully\n"
	   "         before abortion\n", instance->host_no);
    return 0;
}

int wd33c93_reset(struct Scsi_Host *instance)
{
    unsigned long flags;

    printk("scsi%d: reset\n", instance->host_no);
    save_flags(flags);
    cli();
    ((struct WD33C93_hostdata *)instance->hostdata)->dma_stop(NULL);
    reset_WD33C93(instance);
    restore_flags(flags);
    return 0;
}

void wd33c93_init (struct Scsi_Host *instance, wd33c93_regs *regs,
		   dma_setup_t setup, dma_stop_t stop,
		   unsigned long xfermask, int clock_freq)
{
    struct WD33C93_hostdata *hostdata = (struct WD33C93_hostdata *)
	    instance->hostdata;
    int i;

    hostdata->regp = regs;
    hostdata->pio = 0;
    hostdata->clock_freq = clock_freq;
    hostdata->xfermask = xfermask;
    hostdata->dma_setup = setup;
    hostdata->dma_stop = stop;

    for (i = 0; i < 8; ++i)
	    hostdata->busy[i] = 0;
    hostdata->running = 0;
    hostdata->connected = NULL;
    hostdata->issue_queue = NULL;
    hostdata->disconnected_queue = NULL;
    cli();
    reset_WD33C93(instance);
    sti();
}
