/*
** head.S -- This file contains the initial boot code for the the
**	     Amiga Linux kernel.
**
** Copyright 1993 by Hamish Macdonald
**
** This file is subject to the terms and conditions of the GNU General Public
** License.  See the file README.legal in the main directory of this archive
** for more details.
**
*/

#include "assyms.h"

/*
 * Linux startup code.
 *
 * At this point, the boot loader has:
 * Disabled interrupts
 * Disabled caches
 * Put us in supervisor state.
 *
 * The kernel setup code takes the following steps:
 *   Raise interrupt level
 *   Set up initial kernel memory mapping.
 *	This sets up a mapping of the 4M of memory the kernel
 *	is located in.	It also does a mapping of any initial
 *	machine specific areas.
 *   Enable cache memories
 *   Jump to kernel startup
 *
 * Register d6 contains the CPU flags from the boot_info information
 * for most of this file.
 */

.text
.globl _start, _krt, _kpt, _availmem

	.even
_start:
/*
 * raise interrupt level with MASTER bit set, copy isp to msp
 */
	movew	#0x3700,sr
	movec	isp,d0
	movel	d0,sp

/*
 * Initialize serial port
 */
	bsr Lserial_init

	moveq	#13,d7
	bsr	Lserial_putc
	moveq	#10,d7
	bsr	Lserial_putc
	moveq	#65,d7
	bsr	Lserial_putc

/*
 * Record the CPU type.
 */
	lea	pc@(_boot_info),a0
	movel	a0@(LBI_CPU),d6

/*
 * Get address at end of kernel code/data/bss and
 * mask off at a page boundary.
 */
	lea	pc@(_end),a0
	movel	a0,d0
	addl	#0x00000fff,d0
	andl	#0xfffff000,d0
	movel	d0,a6

	moveq	#66,d7
	bsr	Lserial_putc

/*
 * allocate the kernel root table and initialize it.
 */
	movel	a6,a5
	addw	#LPAGE_SIZE,a6

	/* clear the root table */
	movel	a5,a0
	moveq	#0,d0
	moveq	#127,d1
1:	movel	d0,a0@+
	dbra	d1,1b

	/*
	 * Initialize root table descriptor pointing to the kernel pointer
	 * table.  The pointer table is situated in memory right after
	 * the root table.
	 */
	lea	a5@(512),a0
	addql	#2,a0
	movel	#LKSTART_ADDR,d0
	moveq	#25,d1
	lsrl	d1,d0
	movel	a0,a5@(d0:l:4)

	moveq	#67,d7
	bsr	Lserial_putc

/*
 * Initialize the pointer tables referred to above.  They either point
 * to page tables in the case of the 68040 or contain early
 * termination page descriptors in the case of the 68851 or 68030.
 *
 * Each pointer table entry points to a 64 entry page table.  16 of these
 * page tables are grouped to form a single 1024 entry page table which
 * fits in a single 4096 byte page.
 *
 * Some register usages:
 *    a0 -> pointer table descriptor address
 *    a1 -> pointer table descriptor
 *    d1 -> counter
 *    d2 -> pointer table descriptor increment (varies according to CPU)
 */

	/* clear the kernel pointer table */
	lea	a5@(512),a0
	moveq	#0,d0
	moveq	#127,d1
1:	movel	d0,a0@+
	dbra	d1,1b

	lea	a5@(512),a0
	moveq	#15,d1

	/*
	 * base value of pointer table descriptor is either
	 * the address of the first page table (68040)
	 * or the base address of physical memory (68030).
	 */
	btst	#2,d6
	bnes	1f

	/* 680[23]0 */
	lea	pc@(_start),a1  /* base address */
	addql	#1,a1		/* descriptor type */
	movel	#0x40000,d2	/* increment */
	bras	2f

1:	/* 68040 */
	movel	a6,a1		/* base address */
	addw	#LPAGE_SIZE,a6	/* allocate the page table */
	lea	pc@(_kpt),a3
	movel	a1,a3@		/* save address of page table */
	addql	#2,a1		/* descriptor type */
	movel	#256,d2 	/* increment */

2:	movel	a1,a0@+
	addl	d2,a1
	dbra	d1,2b

	moveq	#68,d7
	bsr	Lserial_putc

/*
 * If we are running on a 68040, we have a kernel page table and
 * must initialize it.	Make the entries point to the first
 * 4M of physical memory (the memory we are residing in).
 * Set the cache mode bits to Cacheable, Copyback.  Set the Global bits
 * in the descriptors also.
 */

	btst	#2,d6
	beqs	Lnot040

	moveq	#70,d7
	bsr	Lserial_putc

	movel	pc@(_kpt),a0
	lea	pc@(_start),a1
	addw	#0x421,a1
	movel	#1023,d1
	movel	#LPAGE_SIZE,d2

1:	movel	a1,a0@+
	addl	d2,a1
	dbra	d1,1b

	/*
	 * on the 68040, pages used to hold mmu tables should
	 * be initialized as noncachable/serialized.
	 * Do this for the root table page (which also contains
	 * all pointer tables utilitized thus far) and the
	 * kernel page table.
	 */
	lea	pc@(_start),a0
	movel	a5,d0		/* address of root table */
	subl	a0,d0		/* determine offset of root table page */
	moveq	#12,d1		/* determine offset into kernel page table */
	lsrl	d1,d0		/* i.e. page number of the address offset */
	movel	pc@(_kpt),a0
	lea	a0@(d0:l:4),a0
	movel	a0@,d1
	bclr	#5,d1
	bset	#6,d1
	movel	d1,a0@

	movel	a0@(4),d1       /* do the same for the kernel page table */
	bclr	#5,d1		/* the kernel page table resides in the  */
	bset	#6,d1		/* page after the page containing the	 */
	movel	d1,a0@(4)       /* root table                            */

Lnot040:
/*
 * Do any machine specific page table initializations.
 */
	lea	pc@(_boot_info),a0
	moveq	#LMACH_AMIGA,d0
	cmpl	a0@(LBI_MACH),d0
	bne	Lnotami

/*
 * On the Amiga:
 * Setup "transparent" mapping of first 16M, using early termination
 * page descriptors for the 68030, and proper page tables for the 68040.
 * Set this area as non-cacheable.
 */

	moveq	#72,d7
	bsr	Lserial_putc

	btst	#2,d6
	bnes	Lspami68040

	/*
	 * for the 68030, just setup an identity translation for the
	 * the first 32M of virtual address space by using an early
	 * termination page descriptor pointing to the first 32M of
	 * the physical address space.
	 */

	moveq	#73,d7
	bsr	Lserial_putc

	moveq	#0x41,d0
	movel	d0,a5@

	bra	Lmapphys

Lspami68040:

	/*
	 * for the 68040, use another pointer table, and allocate 4 more
	 * page tables.  Initialize the pointer table to point to the
	 * page tables.  Then initialize the page tables to point to
	 * the first 16M of memory, with no caching (noncachable/serialized).
	 */

	/* clear the amiga pointer table */
	lea	a5@(1024),a0
	moveq	#0,d0
	moveq	#127,d1
1:	movel	d0,a0@+
	dbra	d1,1b

	/* allocate 4 page tables */
	movel	a6,a3
	addw	#LPAGE_SIZE*4,a6

	/* initialize the pointer table */
	lea	a5@(1024),a0
	movel	a3,a1
	addql	#2,a1		/* base descriptor */
	movel	#256,d2 	/* increment */
	moveq	#63,d1

1:	movel	a1,a0@+
	addl	d2,a1
	dbra	d1,1b

	/*
	 * initialize the page tables
	 * descriptor bits include noncachable/serialized and global bits.
	 */
	movel	a3,a0
	movew	#0x441,a1
	movel	#LPAGE_SIZE,d2
	movel	#1024*4,d1

1:	movel	a1,a0@+
	addl	d2,a1
	dbra	d1,1b

	/*
	 * Finally, since we just allocated 4 page table, make sure that
	 * the virtual mapping of the 4 page tables indicates
	 * noncachable/serialized.
	 */
	movel	a3,d0		/* ami page table start address */
	lea	pc@(_start),a0
	subl	a0,d0		/* determine offset of root table page */
	moveq	#12,d1		/* determine offset into kernel page table */
	lsrl	d1,d0
	movel	pc@(_kpt),a0
	movel	#3,d2
1:	lea	a0@(d0:l:4),a1
	movel	a1@,d1
	bclr	#5,d1
	bset	#6,d1
	movel	d1,a1@
	addql	#1,d0
	dbra	d2,1b

	bra	Lmapphys


Lnotami:	/* other machines specific mappings go here! */

/*
 * Setup a transparent mapping of the physical memory we are executing in.
 *
 * Only do this if the physical memory is not in the first 16M Meg, or not on
 * an Amiga since the first 16M is already identity mapped on the Amiga.
 */
Lmapphys:
	moveq	#74,d7
	bsr	Lserial_putc

	lea	pc@(_boot_info),a0
	cmpl	#LMACH_AMIGA,a0@(LBI_MACH)
	bnes	Lnophys 	/* other machines will probably have
				 * to put in code and jump to it here
				 */

	/* 68040? */
	btst	#2,d6
	beqs	1f

	movel	#0x01000000,d0	/* limit for 68040 is 16M (16M mapped for
				 * Amiga
				 */
	bras	2f

1:	movel	#0x02000000,d0	/* limit for 68030 is 32M (32M was mapped
				 * physically using an early term page desc
				 */

2:
	lea	pc@(_start),a0  /* get address of start of memory */
	cmpl	a0,d0		/* compare limit to start */
	bhi	Lnophys

	/* 68040? */
	btst	#2,d6
	bnes	1f

	/*
	 * for 68030, just setup an early termination page descriptor
	 * in the root table.  a0 contains the start address of the
	 * kernel.
	 */
	lea	pc@(_start),a0
	movel	a0,d2
	moveq	#25,d3		/* find appropriate index in root table */
	lsrl	d3,d2
	movel	d2,d0
	lsll	d3,d0		/* convert address into descriptor */
	addql	#1,d0
	movel	d0,a5@(d2:l:4)  /* store mapping */
	bras	Lnophys

1:	/*
	 * For the 68040, we will use the transparent translation
	 * registers to identity map the 16M chunk that contains
	 * the physical memory.
	 */
	lea	pc@(_start),a0
	movel	a0,d2
	andl	#0xff000000,d2		/* logical address base */
	orw	#0xc040,d2		/* add in magic bits */
	.long	0x4e7b2004		/* movec d2,ittr0 */
	.long	0x4e7b2006		/* movec d2,dttr0 */

Lnophys:

/*
 * Setup Supervisor Root Pointer register to point to page directory,
 * setup translation register contents and enable translation.
 */
	btst	#2,d6
	bnes	Lmmu68040

	moveq	#75,d7
	bsr	Lserial_putc

	lea	pc@(Lmmu),a0
	movel	#0x80000002,a0@   /* no limit, 4byte descriptors */
	movel	a5,a0@(4)
	pmove	a0@,srp

	/*
	 * enable,super root enable,4096 byte pages,7 bit root index,
	 * 7 bit pointer index, 6 bit page table index.
	 */
	movel	#0x82c07760,a0@
	pmove	a0@,tc

	bras	1f

Lmmu68040:
	moveq	#76,d7
	bsr	Lserial_putc

	.word	0xf518		/* pflusha	*/
	.long	0x4e7bd807	/* movec a5,srp */
	movel	#0x8000,d0	/* enable translation, 4096 byte pages */
	.long	0x4e7b0003	/* movec d0,tc	*/

1:

/*
 * Fixup the addresses for the kernel pointer table and availmem.
 * Convert them from physical addresses to virtual addresses.
 */
	moveq	#77,d7
	bsr	Lserial_putc

	movel	#LKSTART_ADDR,d0
	lea	pc@(_start),a0

	movel	_kpt,d1
	subl	a0,d1
	addl	d0,d1
	movel	d1,_kpt

	/*
	 * do the same conversion on the first available memory
	 * address (in a6).
	 */
	subl	a0,a6
	addl	d0,a6
	movel	a6,_availmem	/* first available memory address */

	/* convert and save the root table address in memory */
	subl	a0,a5
	addl	d0,a5
	movel	a5,_krt

	btst	#2,d6
	beqs	1f
	lea	Lvirt,a0
	moveq	#2,d0
	movec	d0,dfc
	.word	0xf568		/* ptestr a0@ */
	.long	0x4e7a0805	/* movec mmusr,d0 */
	moveq	#32,d7
	bsr	Lserial_putc
	movel	d0,d7
	bsr	Lserial_putnum
	moveq	#13,d7
	bsr	Lserial_putc
	moveq	#10,d7
	bsr	Lserial_putc

1:
	/*
	 * Record starting address and jump into the virtual address space.
	 */
	lea	pc@(_start),a0
	jmp	Lvirt

Lvirt:
	moveq	#78,d7
	bsr	Lserial_putc

	/*
	 * Clean up the temporary physical mapping.
	 */
	btst	#2,d6
	bnes	Loff040

	/* clean up physical identity mapping for 68020/68030 */
	movel	a0,d2		/* a0 contains physical start address */
	moveq	#25,d3		/* find appropriate index in root table */
	lsrl	d3,d2
	moveq	#0,d0
	movel	d0,a5@(d2:l:4)  /* clear descriptor */
	pflusha 		/* flush address translation cache */
	bras	1f

Loff040:
	/* turn off transparent translation registers for '040 */
	moveq	#0,d0
	.long	0x4e7b0004	/* movec d0,ittr0 */
	.long	0x4e7b0006	/* movec d0,dttr0 */

1:

	moveq	#79,d7
	bsr	Lserial_putc


/*
 * Enable caches
 */
	btst	#2,d6
	bnes	Lcache68040

	movel	#0x00000919,d0
	movec	d0,cacr
	bras	1f

Lcache68040:
	.word	0xf4d8	     /* CINVA I/D */
	movel	#0x80008000,d0
	/* don't enable cache on 68040 until MMU stuff works */
/*	movec	d0,cacr */

1:

/*
 * Setup initial stack pointer
 */
	leal	_user_stack+LPAGE_SIZE,sp

/* jump to memory management initialization */

	moveq	#0,d0
	movel	d0,sp@- 	/* parameters to main */
	movel	d0,sp@-
	movel	d0,sp@-
	bsr	_start_kernel


/*
 * Serial port output support.
 */
LSERPER      = 0xdff032
LSERDAT      = 0xdff030
LSERDATR     = 0xdff018
LNTSC_PERIOD = 371
LPAL_PERIOD  = 368
LNTSC_ECLOCK = 7159090
LSERIAL_CNTRL = 0xbfd000
LSERIAL_DTR   = 7


/*
 * Initialize serial port hardware for 9600/8/1
 */
Lserial_init:
	lea	pc@(_boot_info),a0
	cmpil	#LMACH_AMIGA,a0@(LBI_MACH)
	bnes	1f
	bclr	#LSERIAL_DTR,LSERIAL_CNTRL
	movew	#LNTSC_PERIOD,LSERPER
	cmpl	#LNTSC_ECLOCK,a0@(LBI_AMIGA_ECLK)
	beqs	1f
	movew	#LPAL_PERIOD,LSERPER
1:	rts

/*
 * Output character in d7 on serial port.
 * d7 unchanged.
 */
Lserial_putc:
	movel	a0,sp@-
	lea	pc@(_boot_info),a0
	cmpil	#LMACH_AMIGA,a0@(LBI_MACH)
	bnes	2f
	andw	#0x00ff,d7
	oriw	#0x0100,d7
	movew	d7,LSERDAT
1:	movew	LSERDATR,d7
	andw	#0x2000,d7
	beq	1b
2:	movel	sp@+,a0
	rts

/*
 * Output string pointed to by a0 to serial port.
 * a0 trashed.
 */
Lserial_puts:
	movel	d7,sp@-
1:	moveb	a0@+,d7
	beqs	2f
	bsrs	Lserial_putc
	bras	1b
2:	movel	sp@+,d7
	rts

/*
 * Output number in d7 in hex notation on serial port.
 * d0-d2 trashed.
 * d7 trashed.
 */
Lserial_putnum:
	moveml	d0-d2/d7,sp@-
	movel	d7,d1
	moveq	#4,d0
	moveq	#7,d2
L1:	roll	d0,d1
	moveb	d1,d7
	andb	#0x0f,d7
	cmpb	#0x0a,d7
	bccs	1f
	addb	#48,d7
	bras	2f
1:	addb	#55,d7
2:	bsrs	Lserial_putc
	dbra	d2,L1
	moveq	#32,d7
	bsrs	Lserial_putc
	moveml	sp@+,d0-d2/d7
	rts
.globl showtest
showtest:
	moveml	a1/d7,sp@-
	moveq	#65,d7
	bsrs	Lserial_putc
	moveq	#61,d7
	bsrs	Lserial_putc
	movel	a0,d7
	bsrs	Lserial_putnum

	ptestr	#5,a0@,#7,a1

	moveq	#68,d7
	bsr	Lserial_putc
	moveq	#65,d7
	bsr	Lserial_putc
	moveq	#61,d7
	bsr	Lserial_putc

	movel	a1,d7
	bsrs	Lserial_putnum

	moveq	#68,d7
	bsr	Lserial_putc
	moveq	#61,d7
	bsr	Lserial_putc
	movel	a1@,d7
	bsrs	Lserial_putnum

	moveq	#83,d7
	bsr	Lserial_putc
	moveq	#61,d7
	bsr	Lserial_putc

	lea	pc@(Lmmu),a1
	pmove	psr,a1@
	clrl	d7
	movew	a1@,d7
	bsr	Lserial_putnum

	moveq	#10,d7
	bsr	Lserial_putc
	moveq	#13,d7
	bsr	Lserial_putc

	moveml	sp@+,a1/d7
	rts

	.data
Lmmu:	.quad 0
_krt:	.long 0
_kpt:	.long 0
_availmem:
	.long 0
